#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <asm/types.h>
#include <linux/videodev2.h>
#include <sys/ioctl.h>
#include <fltk/Button.h>
#include <fltk/Widget.h>
#include <fltk/Window.h>
#include <fltk/run.h>

#include "jpeg_compress.h"
#include "kohn_avi.h"
#include "VideoWindow.h"

using namespace fltk;

static unsigned char left_buffer[VIDEO_WIDTH*VIDEO_HEIGHT*3] = { 0 };
static unsigned char right_buffer[VIDEO_WIDTH*VIDEO_HEIGHT*3] = { 0 };
static int left_fd;
static int right_fd;
static struct kavi_t *kavi=NULL;

void timeout_callback(void *param)
{
  VideoWindow *window=(VideoWindow *)param;
  //unsigned char buffer[640*480*2];
  int len;
  int c;
  int n;

  if (left_fd>0)
  {
    len=640*480*2;
    c=0;
    while(c<len)
    {
      n=read(left_fd, left_buffer+c, len-c);
      if (n<0)
      {
        printf("v4l2 read()=%d error.  errno=%d  buffer=%d/%d\n",n,errno,c,len);
        break;
      }

      c=c+n;
    }
  }

  if (right_fd>0)
  {
    len=640*480*2;
    c=0;
    while(c<len)
    {
      n=read(right_fd, right_buffer+c, len-c);
      if (n<0)
      {
        printf("v4l2 read()=%d error.  errno=%d  buffer=%d/%d\n",n,errno,c,len);
        break;
      }

      c=c+n;
    }
  }

  //convert_yuyv(window->video_buffer, buffer, 640, 480);

  window->add_frame(left_buffer, right_buffer);
  window->redraw();

  repeat_timeout(0.033, timeout_callback, window);
}

void snapshot_callback(Widget *w, void *data)
{
  VideoWindow *window=(VideoWindow *)data;
  unsigned char jpeg_buffer[128*1024];

  int len=jpeg_compress(window->getBuffer(), VIDEO_WIDTH*VIDEO_HEIGHT*3, jpeg_buffer, 128*1024, VIDEO_WIDTH, VIDEO_HEIGHT, 3, 90);

  FILE *fp=fopen("out.jpeg", "wb");
  fwrite(jpeg_buffer, 1, len, fp);
  fclose(fp);

}

void open_callback(Widget *w, void *data)
{
  if (kavi!=NULL) return;
  kavi=kavi_open((char *)"out.avi", 640, 480, (char *)"MJPG", 15, NULL);
}

void add_callback(Widget *w, void *data)
{
  if (kavi==NULL) return;
  VideoWindow *window=(VideoWindow *)data;
  unsigned char jpeg_buffer[128*1024];

  int len=jpeg_compress(window->getBuffer(), VIDEO_WIDTH*VIDEO_HEIGHT*3, jpeg_buffer, 128*1024, VIDEO_WIDTH, VIDEO_HEIGHT, 3, 90);

  kavi_add_frame(kavi, jpeg_buffer, len);
}

void close_callback(Widget *w, void *data)
{
  if (kavi==NULL) return;
  kavi_close(kavi);
  kavi=NULL;
}

int main(int argc, char *argv[])
{

  VideoWindow *window=new VideoWindow(800,480);
  window->begin();

  Button snapshot_button(655,20,130,40, "Snapshot");
  Button open_button(655,200,130,40, "Open AVI");
  Button add_button(655,250,130,40, "Add Frame");
  Button close_button(655,300,130,40, "Close AVI");

  snapshot_button.callback(snapshot_callback, window);
  open_button.callback(open_callback, NULL);
  add_button.callback(add_callback, window);
  close_button.callback(close_callback, NULL);

  left_fd=open("/dev/video1", O_RDWR, 0);
  right_fd=open("/dev/video2", O_RDWR, 0);

  if (left_fd<0)
  {
    printf("Could not open video device\n");
  }
    else
  {
    struct v4l2_capability vid_cap;
    struct v4l2_format vid_fmt;

    if (ioctl(left_fd, VIDIOC_QUERYCAP, &vid_cap)==-1)
    {
      printf("ioctl error: VIDIOC_QUERYCAP  errno=%d\n",errno);
      return -1;
    }

    printf("      driver: %16s\n",vid_cap.driver);
    printf("        card: %32s\n",vid_cap.card);
    printf("    bus_info: %32s\n",vid_cap.bus_info);
    printf("     version: %d\n",vid_cap.version);
    printf("capabilities: 0x%08x\n",vid_cap.capabilities);
    printf("\n");

    if ((vid_cap.capabilities&V4L2_CAP_VIDEO_CAPTURE)==0)
    {
      close(left_fd);
      printf("This device doesn't support frame capture\n");
      return -3;
    }

    vid_fmt.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
    vid_fmt.fmt.pix.width=640;
    vid_fmt.fmt.pix.height=480;
    //vid_fmt.fmt.pix.pixelformat=V4L2_PIX_FMT_RGB24;
    vid_fmt.fmt.pix.pixelformat=V4L2_PIX_FMT_YUYV;
    //vid_fmt.fmt.pix.pixelformat=V4L2_PIX_FMT_BGR24;
    // vid_fmt.fmt.pix.pixelformat=V4L2_PIX_FMT_RGB32;
    //vid_fmt.fmt.pix.pixelformat=V4L2_PIX_FMT_YUV420;
    vid_fmt.fmt.pix.field=V4L2_FIELD_ANY;

    if (ioctl(left_fd, VIDIOC_S_FMT, &vid_fmt)==-1)
    {
      close(left_fd);
      printf("ioctl error: VIDIOC_S_FMT errno=%d\n",errno);
      return -1;
    }

    //memset(&vid_fmt, 0, sizeof(vid_fmt));

    if (ioctl(left_fd, VIDIOC_G_FMT, &vid_fmt)==-1)
    {
      close(left_fd);
      printf("ioctl error: VIDIOC_G_FMT errno=%d\n",errno);
      return -1;
    }

    printf(" width: %d\n", vid_fmt.fmt.pix.width);
    printf("height: %d\n", vid_fmt.fmt.pix.height);
    printf("format: %x %c%c%c%c\n",
      vid_fmt.fmt.pix.pixelformat,
      vid_fmt.fmt.pix.pixelformat&0xff,
      (vid_fmt.fmt.pix.pixelformat>>8)&0xff,
      (vid_fmt.fmt.pix.pixelformat>>16)&0xff,
      (vid_fmt.fmt.pix.pixelformat>>24)&0xff
    );

    //window->video_buffer=(unsigned char *)malloc(640*480*3);

    //add_fd(left_fd, READ, (FileHandler)video_callback, window);
  }

  if (right_fd<0)
  {
    printf("Could not open video device\n");
  }
    else
  {
    struct v4l2_capability vid_cap;
    struct v4l2_format vid_fmt;

    if (ioctl(right_fd, VIDIOC_QUERYCAP, &vid_cap)==-1)
    {
      printf("ioctl error: VIDIOC_QUERYCAP  errno=%d\n",errno);
      return -1;
    }

    printf("      driver: %16s\n",vid_cap.driver);
    printf("        card: %32s\n",vid_cap.card);
    printf("    bus_info: %32s\n",vid_cap.bus_info);
    printf("     version: %d\n",vid_cap.version);
    printf("capabilities: 0x%08x\n",vid_cap.capabilities);
    printf("\n");

    if ((vid_cap.capabilities&V4L2_CAP_VIDEO_CAPTURE)==0)
    {
      close(right_fd);
      printf("This device doesn't support frame capture\n");
      return -3;
    }

    vid_fmt.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
    vid_fmt.fmt.pix.width=640;
    vid_fmt.fmt.pix.height=480;
    //vid_fmt.fmt.pix.pixelformat=V4L2_PIX_FMT_RGB24;
    vid_fmt.fmt.pix.pixelformat=V4L2_PIX_FMT_YUYV;
    //vid_fmt.fmt.pix.pixelformat=V4L2_PIX_FMT_BGR24;
    // vid_fmt.fmt.pix.pixelformat=V4L2_PIX_FMT_RGB32;
    //vid_fmt.fmt.pix.pixelformat=V4L2_PIX_FMT_YUV420;
    vid_fmt.fmt.pix.field=V4L2_FIELD_ANY;

    if (ioctl(right_fd, VIDIOC_S_FMT, &vid_fmt)==-1)
    {
      close(right_fd);
      printf("ioctl error: VIDIOC_S_FMT errno=%d\n",errno);
      return -1;
    }

    //memset(&vid_fmt, 0, sizeof(vid_fmt));

    if (ioctl(right_fd, VIDIOC_G_FMT, &vid_fmt)==-1)
    {
      close(right_fd);
      printf("ioctl error: VIDIOC_G_FMT errno=%d\n",errno);
      return -1;
    }

    printf(" width: %d\n", vid_fmt.fmt.pix.width);
    printf("height: %d\n", vid_fmt.fmt.pix.height);
    printf("format: %x %c%c%c%c\n",
      vid_fmt.fmt.pix.pixelformat,
      vid_fmt.fmt.pix.pixelformat&0xff,
      (vid_fmt.fmt.pix.pixelformat>>8)&0xff,
      (vid_fmt.fmt.pix.pixelformat>>16)&0xff,
      (vid_fmt.fmt.pix.pixelformat>>24)&0xff
    );

    //window->video_buffer=(unsigned char *)malloc(640*480*3);

    //add_fd(right_fd, READ, (FileHandler)video_callback, window);
  }

  add_timeout(0.033, timeout_callback, window);

  window->end();
  window->show(argc, argv);

  return run();
}


